try:
    from urllib.parse import unquote
except ImportError:
    from urllib import unquote

import socket
import email
from zokket.tcp import TCPSocket


class WSGIRequestHandler(object):
    def __init__(self, server, sock, data):
        self.server = server
        self.version = self.server.request_version
        self.socket = sock

        self.handle_data(data)

    def start_response(self, status, headers):
        self.status = status
        self.response_headers = headers

    def handle_data(self, data):
        lines = data.split("\r\n")[:-2]
        request_line = lines.pop(0)
        request_words = request_line.split()

        if len(request_words) == 3:
            [command, path, version] = request_words

            if version[:5] != 'HTTP/':
                self.send_error('400 BAD REQUEST', 'Bad request version ({})'.format(version))
                return
            self.version = version
        elif len(request_words) == 2:
            [command, path] = request_words
        elif not request_words:
            return
        else:
            self.send_error('400 BAD REQUEST', 'Bad request syntax ({})'.format(request_line))
            return

        self.environ = self.server.base_environ.copy()
        self.headers = email.message_from_string("\r\n".join(lines))
        self.environ['SERVER_PROTOCOL'] = version
        self.environ['REQUEST_METHOD'] = command

        if '?' in path:
            path, query = path.split('?', 1)
        else:
            path, query = path, ''

        self.environ['PATH_INFO'] = unquote(path)
        self.environ['QUERY_STRING'] = query
        self.environ['CONTENT_TYPE'] = self.headers.get('Content-Type', '')
        self.environ['CONTENT_LENGTH'] = self.headers.get('Content-Length', '')

        for key, value in self.headers.items():
            key = ('HTTP_' + key.upper().replace('-', '_'))
            if key not in ('HTTP_CONTENT_TYPE', 'HTTP_CONTENT_LENGTH'):
                self.environ[key] = value

        self.environ['REMOTE_HOST'] = self.socket.connected_host()
        self.environ['REMOTE_ADDR'] = self.socket.connected_host()

        self.handle_request()
        self.send_response()

        self.socket.socket.shutdown(socket.SHUT_WR)

        conntype = self.headers.get('Connection', "")
        if conntype.lower() != 'keep-alive':
            self.socket.close()

    def handle_request(self):
        try:
            self.data = self.server.handler(self.environ, self.start_response)
        except Exception:
            self.status = '500 INTERNAL SERVER ERROR'
            self.response_headers = [('Content-type', 'text/plain')]
            self.data = ['Internal Server Error']

    def send_error(self, error, explain):
        self.status = error
        self.data = explain
        self.response_headers = [('Content-type', 'text/plain')]
        self.send_response()
        self.socket.close()

    def send_response(self):
        self.socket.send("{} {}\r\n".format(self.version, self.status))
        self.socket.send("\r\n".join(['{}: {}'.format(*header) for header in self.response_headers]) + "\r\n\r\n")
        self.socket.send("\n".join(self.data) + "\r\n\r\n")


class WSGIServer(object):
    protocol_version = "HTTP/1.0"
    request_version = "HTTP/0.9"

    def __init__(self, handler, host='', port=8082):
        self.handler = handler

        self.socket = TCPSocket(self)
        self.socket.accept(host, port)

        self.base_environ = {}
        self.base_environ['SERVER_NAME'] = socket.getfqdn(self.socket.local_host())
        self.base_environ['SERVER_PORT'] = str(self.socket.local_port())
        self.base_environ['GATEWAY_INTERFACE'] = 'CGI/1.1'
        self.base_environ['REMOTE_HOST'] = ''
        self.base_environ['CONTENT_LENGTH'] = ''
        self.base_environ['SCRIPT_NAME'] = ''

    def socket_did_connect(self, sock, host, port):
        sock.read_until_data = "\r\n\r\n"

    def socket_read_data(self, sock, data):
        WSGIRequestHandler(self, sock, data)



from zokket.tcp import SocketException, TCPSocket
from zokket.udp import UDPSocket
from zokket.timers import Timer
from zokket.runloop import DefaultRunloop, Runloop
from zokket.wsgi import WSGIServer

VERSION = (1, 2, 0)

def get_version():
    version = '{}.{}'.format(VERSION[0], VERSION[1])
    if VERSION[2]:
        version = '{}.{}'.format(version, VERSION[2])
    return version



from PyQt4.QtCore import QObject, QSocketNotifier, SIGNAL, QBasicTimer
from zokket.runloop import DefaultRunloop


class QtRunloop(QObject):
    @classmethod
    def set_default(cls, app=None):
        DefaultRunloop.set(cls)

        if app:
            DefaultRunloop.default().set_application(app)

    def __init__(self):
        super(QtRunloop, self).__init__()
        self.sockets = []
        self.timers = []

    def set_application(self, app):
        self.connect(app, SIGNAL('aboutToQuit()'), self.shutdown)

    def shutdown(self):
        [s.close() for s in self.sockets]

    # timers

    def register_timer(self, timer):
        self.timers.append(timer)

        timer.qtimer = QBasicTimer()
        timer.qtimer.start(timer.interval * 1000, self)

    def unregister_timer(self, timer):
        if hasattr(timer, 'qtimer') and timer.qtimer:
            timer.qtimer.stop()
            timer.qtimer = None

        self.timers.remove(timer)

    def timerEvent(self, event):
        for timer in self.timers:
            if timer.qtimer.timerId() == event.timerId():
                return timer.execute()

    # sockets

    def register_socket(self, socket):
        if socket not in self.sockets:
            self.sockets.append(socket)

        socket.read_notifier = QSocketNotifier(socket.fileno(), QSocketNotifier.Read)
        self.connect(socket.read_notifier, SIGNAL("activated(int)"), self.read_socket)
        socket.read_notifier.setEnabled(socket.readable())

        socket.write_notifier = QSocketNotifier(socket.fileno(), QSocketNotifier.Write)
        self.connect(socket.write_notifier, SIGNAL("activated(int)"), self.write_socket)
        socket.write_notifier.setEnabled(socket.writable())

        #socket.except_notifier = QSocketNotifier(socket.fileno(), QSocketNotifier.Exception)
        #self.connect(socket.except_notifier, SIGNAL("activated(int)"), self.except_event)
        #socket.except_notifier.setEnabled(True)

    def unregister_socket(self, socket):
        if hasattr(self.sockets, 'read_notifier') and self.sockets.read_notifier:
            self.sockets.read_notifier.setEnabled(False)
            self.sockets.read_notifier = None

        if hasattr(self.sockets, 'write_notifier') and self.sockets.write_notifier:
            self.sockets.write_notifier.setEnabled(False)
            self.sockets.write_notifier = None

        if hasattr(self.sockets, 'except_notifier') and self.sockets.except_notifier:
            self.sockets.except_notifier.setEnabled(False)
            self.sockets.except_notifier = None

        self.sockets.remove(socket)

    def update_socket(self, socket):
        if hasattr(socket, 'read_notifier') and socket.read_notifier:
            socket.read_notifier.setEnabled(socket.readable())

        if hasattr(socket, 'write_notifier') and socket.write_notifier:
            socket.write_notifier.setEnabled(socket.writable())

        #if hasattr(socket, 'except_notifier') and socket.except_notifier:
        #    socket.except_notifier.setEnabled(True)

    def read_socket(self, fd):
        for socket in self.sockets:
            if socket.fileno() == fd:
                socket.handle_read_event()
                return

    def write_socket(self, fd):
        for socket in self.sockets:
            if socket.fileno() == fd:
                socket.handle_write_event()
                return

    def except_event(self, fd):
        for socket in self.sockets:
            if socket.fileno() == fd:
                socket.handle_except_event()
                return




__all__ = ['SocketException', 'TCPSocketDelegate', 'TCPSocket']

import socket
import os

try:
    import ssl
except ImportError:
    ssl = None

from zokket.timers import Timer
from zokket.runloop import DefaultRunloop

import errno

class SocketException(Exception):
    pass


class TCPSocketDelegate(object):
    """
    An instance of TCPSocket will call methods on its delegate object upon
    completing certain operations or when it encouters errors.

    All instances of TCPSocket should have a delegate that optionally responds
    to these methods. A delegate should be seen as a connection controller.
    """

    def socket_did_disconnect(self, sock, err=None):
        """
        A socket that was previously connected or listening has been closed.

        Additionally an error on connecting could cause this delegate method
        to be called and the exception will be passed.
        """
        pass

    def socket_did_accept_new_socket(self, sock, new_sock):
        """
        An accept socket accepted a new socket.
        """
        pass

    def socket_wants_runloop_for_new_socket(self, sock, new_sock):
        """
        If this method is not implemented then it will use the current runloop.

        This method can be implemented so the socket can run on a seperate
        thread from the accept socket by returning a different runloop
        running on a different thread.
        """
        return RunLoop()

    def socket_will_connect(self, sock):
        """
        The socket calls this method when it is about to connect to a remote
        socket.

        Return True if the socket should continue to connect to the remote
        socket.
        """
        return True

    def socket_connection_refused(self, sock, host, port):
        """
        The connection was refused
        """
        pass

    def socket_did_connect(self, sock, host, port):
        """
        The socket has connected with host and port.
        """
        pass

    def socket_connection_timeout(self, sock, host, port):
        """
        The socket timedout trying to connect to host and port.
        """
        pass

    def socket_read_data(self, sock, data):
        """
        The socket has received data.
        """
        pass

    def socket_address_in_use(self, sock, host, port):
        """
        This method will be called if the address you tried to
        accept on is already in use.
        """
        pass

    def socket_address_refused(self, sock, host, port):
        """
        This method will be called if the address you tried to
        accept on is refused, this usually means you are trying
        to accept on a port <1024 without root priviledges.
        """
        pass

    def socket_accepting(self, sock, host, port):
        """
        This method will be called when a socket is successfully
        listening on the host and port.
        """
        pass

    def socket_did_secure(self, sock):
        """
        This method will be called when TLS negotiation is
        complete, and the rest of the connection will be
        secure. (Unless sock.stop_tls is called).
        """
        pass


class TCPSocket(object):
    def __init__(self, delegate=None, runloop=None):
        """
        The delegate is an object which follows the TCPSocketDelegate protocol.
        """

        self.delegate = delegate
        self.socket = None
        self.connected = False
        self.accepting = False
        self.accepted = False
        self.closing = False
        self.connect_timeout = None

        self.tls_handshake_stage = None
        self.tls_shutdown = False

        self.read_until_data = None
        self.read_until_length = None
        self.read_buffer = bytes()
        self.buffer_type = 'utf-8'

        self.uploaded_bytes = 0
        self.downloaded_bytes = 0

        self.runloop = runloop

    def __str__(self):
        if self.connected:
            return '{}:{}'.format(self.connected_host(), self.connected_port())
        elif self.accepting:
            return '{}:{}'.format(self.local_host(), self.local_port())

        return ''

    def __repr__(self):
        if self.connected:
            return '<TCPSocket connect({}:{})>'.format(
                self.connected_host(),
                self.connected_port()
            )
        elif self.accepting:
            return '<TCPSocket listen({}:{})>'.format(
                self.local_host(),
                self.local_port()
            )
        elif self.socket != None:
            return '<TCPSocket connecting>'

        return '<TCPSocket>'

    def __del__(self):
        self.close()

    def __nonzero__(self):
        return (self.socket != None) and self.connected or self.accepting

    def __eq__(self, other):
        if self.socket is not None and other == self.fileno():
            return True

        if id(self) == id(other):
            return True

        return False

    # Configuration

    def configure(self):
        self.accepted = True

        self.runloop.register_socket(self)

        if not self.socket_will_connect():
            self.close()

    @property
    def runloop(self):
        if not self._runloop:
            self._runloop = DefaultRunloop.default()

        return self._runloop

    @runloop.setter
    def runloop(self, runloop):
        self._runloop = runloop

    # TLS

    def has_tls(self):
        return ssl is not None

    def start_tls(self, keyfile=None, certfile=None, cert_reqs=ssl.CERT_NONE, \
            ca_certs=None, version=ssl.PROTOCOL_SSLv23):
        if not self.has_tls():
            return False

        if self.socket is None:
            return False

        self.socket = ssl.SSLSocket(self.socket, keyfile=keyfile, \
                                    certfile=certfile, cert_reqs=cert_reqs, \
                                    ca_certs=ca_certs, ssl_version=version, \
                                    server_side=self.accepted, \
                                    do_handshake_on_connect=False)
        self.tls_handshake()

        return True

    def stop_tls(self):
        self.tls_shutdown = True

        try:
            self.socket = self.socket.unwrap()
        except ssl.SSLError as err:
            if err.args[0] == ssl.SSL_ERROR_WANT_READ:
                self.tls_handshake_stage = 1
                self.runloop.update_socket(self)
                return False
            elif err.args[0] == ssl.SSL_ERROR_WANT_WRITE:
                self.tls_handshake_stage = 2
                self.runloop.update_socket(self)
                return False
            else:
                raise

        self.tls_shutdown = False
        self.tls_handshake_stage = None

        if self.closing:
            self.close()
        else:
            self.runloop.update_socket(self)

        return True

    def tls_cipher(self):
        if not self.has_tls():
            return

        if isinstance(self.socket, ssl.SSLSocket):
            return self.socket.cipher()

    def tls_peer_certificate(self, binary_form=False):
        if not self.has_tls():
            return

        if isinstance(self.socket, ssl.SSLSocket):
            return self.socket.getpeercert(binary_form)

    # Connecting

    def connect(self, host, port, timeout=None):
        """
        Try to establish a connection with host and port.

        If a timeout is defined, after this timeout has been reached the
        TCPSocket will stop trying to connect and the socket_connection_timeout
        TCPSocketDelegate method will be called.

        If the socket establishes the connection before the timeout
        socket_did_connect TCPSocketDelegate method will be called.
        """

        #if not self.delegate:
        #    raise SocketException("Attempting to accept without a delegate. Set a delegate first.")

        if self.socket != None:
            raise SocketException("Attempting to accept while connected or accepting connections. Disconnect first.")

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setblocking(0)
        self.connecting_address = (host, port)

        self.runloop.register_socket(self)

        try:
            self.socket.connect((host, port))
        except socket.error as e:
            if e.errno in (errno.EINPROGRESS, errno.EALREADY, errno.EWOULDBLOCK):
                if timeout:
                    self.connect_timeout = Timer(timeout, self.connection_timeout, False, (host, port), runloop=self.runloop)
                return
            return self.close(e)

        self.did_connect()

    def connection_timeout(self, timer):
        self.close()
        self.socket_connection_timeout(*timer.data)

    def did_connect(self):
        self.connected = True
        self.runloop.update_socket(self)

        if self.connect_timeout:
            self.connect_timeout.invalidate()

        # Lets make sure that this connection wasn't refused.

        try:
            self.socket.recv(0)
        except socket.error as e:
            if e.errno == errno.ECONNREFUSED:
                self.socket_connection_refused()
                self.close()
                return

        self.socket_did_connect()

    def accept(self, host='', port=0):
        if not self.delegate:
            raise SocketException("Attempting to accept without a delegate. Set a delegate first.")

        if self.socket != None:
            raise SocketException("Attempting to accept while connected or accepting connections. Disconnect first.")

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setblocking(0)

        self.runloop.register_socket(self)

        try:
            self.socket.bind((host, port))
        except socket.error as e:
            if e.errno == errno.EADDRINUSE:
                self.close()
                self.socket_address_in_use(host, port)
                return
            elif e.errno == errno.EACCES:
                self.close()
                self.socket_address_refused(host, port)
                return
            raise e

        self.socket.listen(5)
        self.accepting = True
        self.runloop.update_socket(self)

        self.socket_accepting(host, port)

    def accept_from_socket(self):
        client, address = self.socket.accept()

        new_sock = self.__class__(self.delegate)
        new_sock.socket = client
        new_sock.socket.setblocking(0)

        self.socket_did_accept_new_socket(new_sock)

        runloop = self.socket_wants_runloop_for_new_socket(new_sock)
        new_sock.runloop = runloop or self.runloop
        new_sock.configure()

    def tls_handshake(self):
        if not self.has_tls():
            return

        try:
            self.socket.do_handshake()
            self.tls_handshake_stage = None
            self.runloop.update_socket(self)

            self.socket_did_secure()
        except ssl.SSLError as err:
            if err.args[0] == ssl.SSL_ERROR_WANT_READ:
                self.tls_handshake_stage = 1
                self.runloop.update_socket(self)
            elif err.args[0] == ssl.SSL_ERROR_WANT_WRITE:
                self.tls_handshake_stage = 2
                self.runloop.update_socket(self)
            else:
                raise

    # Disconnect

    def close(self, err=None):
        """
        Disconnect or stop accepting.
        """

        if (self.connected or self.accepted or self.accepting or err):
            self.socket_did_disconnect(err)

        if self.socket != None:
            if isinstance(self.socket, ssl.SSLSocket):
                if not self.stop_tls():
                    self.closing = True
                    return

            self.runloop.unregister_socket(self)

            self.socket.close()
            self.socket = None

        self.closing = False
        self.connected = False
        self.accepting = False

    # Reading

    def read_data(self, data):
        if hasattr(self.delegate, 'socket_read_data'):
            self.delegate.socket_read_data(self, data)

    def dequeue_buffer(self):
        while self.read_buffer:
            if self.read_until_data != None:
                index = self.read_buffer.find(self.read_until_data)

                if index == -1:
                    break

                read_until_data_length = len(self.read_until_data)
                data = self.read_buffer[:index + read_until_data_length]
                self.read_buffer = self.read_buffer[index + read_until_data_length:]

                yield data
            elif self.read_until_length != None:
                if self.read_until_length > len(self.read_buffer):
                    break

                data = self.read_buffer[:self.read_until_length]
                self.read_buffer = self.read_buffer[self.read_until_length:]

                yield data
            else:
                data = self.read_buffer
                self.read_buffer = bytes()
                yield data

    def bytes_availible(self):
        if self.socket == None:
            return

        try:
            if isinstance(self.socket, ssl.SSLSocket):
                data = self.socket.read(8192)
            else:
                data = self.socket.recv(8192)
        except socket.error as e:
            if e.errno in (errno.ECONNRESET, errno.ENOTCONN):
                data = None

        if not data:
            self.close()
        else:
            self.read_buffer += data
            self.downloaded_bytes += len(data)

            for data in self.dequeue_buffer():
                if self.buffer_type:
                    data = data.decode(self.buffer_type)

                self.read_data(data)

    # Writing

    def send(self, data):
        if self.tls_handshake_stage is not None:
            return

        self.uploaded_bytes += len(data)

        if not isinstance(data, bytes):
            data = data.encode(self.buffer_type)

        try:
            if isinstance(self.socket, ssl.SSLSocket):
                return self.socket.write(data)
            return self.socket.send(data)
        except socket.error as e:
            if e.errno == errno.EWOULDBLOCK:
                return 0
            raise e

    # Diagnostics

    def fileno(self):
        if self.socket != None:
            return self.socket.fileno()
        return -1

    def connected_host(self):
        if self.socket == None:
            return ''

        try:
            return self.socket.getpeername()[0]
        except socket.error:
            return ''

    def connected_port(self):
        if self.socket == None:
            return 0

        try:
            return self.socket.getpeername()[1]
        except socket.error as e:
            return -1

    def local_host(self):
        if self.socket == None:
            return ''

        try:
            return self.socket.getsockname()[0]
        except socket.error:
            return ''

    def local_port(self):
        if self.socket == None:
            return 0

        try:
            return self.socket.getsockname()[1]
        except socket.error:
            return 0

    # Runloop Callbacks

    def readable(self):
        return self.socket != None

    def writable(self):
        if self.socket != None:
            if self.tls_handshake_stage == 2:
                return True
            if not self.accepting:
                return not self.connected
        return False

    def handle_read_event(self):
        if self.socket == None:
            return

        if self.tls_handshake_stage is not None:
            if self.tls_shutdown:
                self.stop_tls()
            else:
                self.tls_handshake()
        elif self.accepting:
            self.accept_from_socket()
        elif not self.connected:
            self.did_connect()
            self.bytes_availible()
        else:
            self.bytes_availible()

    def handle_write_event(self):
        if self.socket == None:
            return

        if self.tls_handshake_stage is not None:
            if self.tls_shutdown:
                self.stop_tls()
            else:
                self.tls_handshake()
        elif not self.connected and not self.accepting:
            self.did_connect()

    def handle_except_event(self):
        pass

    def socket_did_disconnect(self, err=None):
        if hasattr(self.delegate, 'socket_did_disconnect'):
            self.delegate.socket_did_disconnect(self, err)

    def socket_did_accept_new_socket(self, new_sock):
        if hasattr(self.delegate, 'socket_did_accept_new_socket'):
            self.delegate.socket_did_accept_new_socket(self, new_sock)

    def socket_wants_runloop_for_new_socket(self, new_sock):
        if hasattr(self.delegate, 'socket_wants_runloop_for_new_socket'):
            return self.delegate.socket_wants_runloop_for_new_socket(self, new_sock)

    def socket_will_connect(self):
        if hasattr(self.delegate, 'socket_will_connect'):
            return self.delegate.socket_will_connect(self)
        return True

    def socket_connection_refused(self):
        if hasattr(self.delegate, 'socket_connection_refused'):
            self.delegate.socket_connection_refused(self, self.connecting_address[0], self.connecting_address[1])

    def socket_did_connect(self):
        if hasattr(self.delegate, 'socket_did_connect'):
            self.delegate.socket_did_connect(self, self.connected_host(), self.connected_port())

    def socket_connection_timeout(self, host, port):
        if hasattr(self.delegate, 'socket_connection_timeout'):
            self.delegate.socket_connection_timeout(self, host, port)

    def socket_address_in_use(self, host, port):
        if hasattr(self.delegate, 'socket_address_in_use'):
            self.delegate.socket_address_in_use(self, host, port)

    def socket_address_refused(self, host, port):
        if hasattr(self.delegate, 'socket_address_refused'):
            self.delegate.socket_address_refused(self, host, port)

    def socket_accepting(self, host, port):
        if hasattr(self.delegate, 'socket_accepting'):
            self.delegate.socket_accepting(self, host, port)

    def socket_did_secure(self):
        if hasattr(self.delegate, 'socket_did_secure'):
            self.delegate.socket_did_secure(self)



import select
import threading

class DefaultRunloop(object):
    @classmethod
    def set(cls, rl):
        cls._runloop = rl()

    @classmethod
    def run_in_new_thread(cls):
        return threading.Thread(target=cls.default().run, args=(), kwargs={}).start()

    @classmethod
    def run(cls):
        return cls.default().run()

    @classmethod
    def default(cls):
        if not hasattr(cls, '_runloop'):
            cls.set(SelectRunloop)
        return cls._runloop

    @classmethod
    def abort(cls):
        cls.default().running = False

class BaseRunloop(object):
    @classmethod
    def set_default(cls):
        DefaultRunloop.set(cls)

    def run(self):
        raise NotImplemented

    def register_socket(self, socket):
        return

    def unregister_socket(self, socket):
        return

    def update_socket(self, socket):
        """
        This is called when the state of a socket changes, when
        socket.writable/readable may of changed.
        """
        return

class TimerRunloopMixin(object):
    def __init__(self):
        self.timers = []

    def register_timer(self, timer):
        if timer not in self.timers:
            self.timers.append(timer)

    def unregister_timer(self, timer):
        if timer in self.timers:
            self.timers.remove(timer)

    def run_timers(self):
        [timer.execute() for timer in self.timers if timer.timeout() <= 0.0]

    def timer_timeout(self):
        try:
            runtimes = [timer.timeout() for timer in self.timers]
            runtimes.sort()
            timeout = runtimes.pop(0)
            if timeout < 0.0:
                return 0
            return timeout
        except:
            return 180

class Runloop(BaseRunloop, TimerRunloopMixin):
    def __init__(self):
        super(Runloop, self).__init__()
        self.sockets = []
        self.running = False

    def run(self):
        self.running = True

        try:
            while self.running:
                self.run_network()
                self.run_timers()
        except (KeyboardInterrupt, SystemExit):
            self.running = False

        self.shutdown()

        while len(self.sockets):
            self.run_network()

    def run_network(self):
        raise NotImplemented

    def shutdown(self):
        [s.close() for s in self.sockets]

    def register_socket(self, socket):
        if socket not in self.sockets:
            self.sockets.append(socket)

    def unregister_socket(self, socket):
        if socket in self.sockets:
            self.sockets.remove(socket)

class SelectRunloop(Runloop):
    def run_network(self):
        r = filter(lambda x: x.readable(), self.sockets)
        w = filter(lambda x: x.writable(), self.sockets)
        e = filter(lambda x: x.socket != None, self.sockets)

        (rlist, wlist, xlist) = select.select(r, w, e, self.timer_timeout())

        [s.handle_except_event() for s in xlist]
        [s.handle_read_event() for s in rlist]
        [s.handle_write_event() for s in wlist]


class PollRunloop(Runloop):
    def __init__(self):
        super(PollRunloop, self).__init__()
        self.poll = select.poll()

    def socket_for_fd(self, fd):
        for socket in self.sockets:
            if socket == fd:
                return socket

    def run_network(self):
        for fd, flag in self.poll.poll():
            socket = self.socket_for_fd(fd)
            if socket is None:
                continue

            if flag & (select.POLLIN | select.POLLPRI):
                socket.handle_read_event()
            elif flag & select.POLLOUT:
                socket.handle_write_event()
            elif flag & select.POLLERR:
                socket.handle_except_event()

    def eventmask(self, socket):
        mask = select.POLLERR | select.POLLHUP | select.POLLPRI

        if socket.writable():
            mask = select.POLLOUT | mask

        if socket.readable():
            mask = select.POLLIN | mask

        return mask

    def register_socket(self, socket):
        super(PollRunloop, self).register_socket(socket)
        self.poll.register(socket, self.eventmask(socket))

    def update_socket(self, socket):
        super(PollRunloop, self).update_socket(socket)
        self.poll.modify(socket, self.eventmask(socket))

    def unregister_socket(self, socket):
        super(PollRunloop, self).unregister_socket(socket)
        self.poll.unregister(socket)



#!/usr/bin/env python2

from distutils.core import setup
from zokket import get_version

kwargs = {
    'name': 'zokket',
    'version': get_version(),
    'license': 'BSD',
    'description': 'zokket is a asynchronous socket networking library for python',
    'author': 'Kyle Fuller',
    'author_email': 'inbox@kylefuller.co.uk',
    'packages': ['zokket'],
    'download_url': 'http://github.com/kylef/zokket/zipball/{}'.format(get_version()),
}

setup(**kwargs)



import socket

from zokket.runloop import DefaultRunloop


class UDPSocketDelegate(object):
    def udp_socket_read_data(self, sock, host, port, data):
        pass


class UDPSocket(object):
    def __init__(self, delegate=None, runloop=None):
        self.delegate = delegate

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.setblocking(0)

        self.uploaded_bytes = 0
        self.downloaded_bytes = 0

        self.runloop = runloop
        self.runloop.register_socket(self)

    def __eq__(self, other):
        if self.socket is not None and other == self.fileno():
            return True

        if id(self) == id(other):
            return True

        return False


    @property
    def runloop(self):
        if not self._runloop:
            self._runloop = DefaultRunloop.default()

        return self._runloop

    @runloop.setter
    def runloop(self, runloop):
        self._runloop = runloop

    # Connecting / Writing

    def bind(self, host='', port=0):
        self.socket.bind((host, port))

    def send(self, host, port, data):
        self.uploaded_bytes += len(data)
        self.socket.sendto(data, (host, port))

    def close(self):
        if self.socket:
            self.runloop.unregister_socket(self)

            self.socket.close()
            self.socket = None

    # Diagnostics

    def fileno(self):
        if self.socket != None:
            return self.socket.fileno()
        return -1

    # Runloop Callbacks

    def readable(self):
        return self.socket != None

    def writable(self):
        return False

    def handle_read_event(self):
        data, addr = self.socket.recvfrom(65565)
        self.downloaded_bytes += len(data)

        if hasattr(self.delegate, 'udp_socket_read_data'):
            self.delegate.udp_socket_read_data(self, addr[0], addr[1], data)

    def handle_write_event(self):
        pass

    def handle_except_event(self):
        pass



import time
from zokket.runloop import DefaultRunloop


class Timer(object):
    def __init__(self, interval, callback, repeat=False, data=None, runloop=None):
        self.interval = interval
        self.callback = callback
        self.repeat = repeat
        self.data = data
        self.runloop = None
        self.update_timeout()
        self.runloop = runloop

        self.runloop.register_timer(self)

    @property
    def runloop(self):
        if not self._runloop:
            self._runloop = DefaultRunloop.default()

        return self._runloop

    @runloop.setter
    def runloop(self, runloop):
        self._runloop = runloop

    def update_timeout(self):
        self.fire_at = time.time() + self.interval

    def timeout(self):
        return self.fire_at - time.time()

    def fire(self):
        self.callback(self)

    def execute(self):
        self.fire()

        if self.repeat:
            self.update_timeout()
        else:
            self.invalidate()

    def invalidate(self):
        self.interval = 0
        self.repeat = False
        self.fite_at = None

        self.runloop.unregister_timer(self)
